﻿using System;
using System.Collections.Generic;
using System.Linq;
using SimulacniProstredi.Grafy;
using System.Drawing;
using System.Drawing.Drawing2D;
using SimulacniProstredi.Sceny;
using System.Text;

namespace SimulacniProstredi.Algoritmy
{
    class AStar
    {
        List<Vertex<VertexData_State_Astar, EdgeData_Heuristic>> AllNodesList;
        StateSpace<VertexData_State_Astar, EdgeData_Heuristic> Graph = new StateSpace<VertexData_State_Astar, EdgeData_Heuristic>();
        VertexData_State_Astar Start, Goal;
        Scene ThisScene;
        List<Vertex<VertexData_State_Astar, EdgeData_Heuristic>> Solution = new List<Vertex<VertexData_State_Astar, EdgeData_Heuristic>>();
        public int NumberOfMathOperations = 0;
        private bool IsInList(Vertex<VertexData_State_Astar, EdgeData_Heuristic> Vertex, List<Vertex<VertexData_State_Astar, EdgeData_Heuristic>> Vertexes)
        {
            foreach (Vertex<VertexData_State_Astar, EdgeData_Heuristic> v in Vertexes)
            {
               if (Vertex.VertexData.Equals(v.VertexData)) { return true; }
            }
            return false;
        }

        public AStar(Scene InScene) 
        {
            ThisScene = InScene;
            Point instart= ThisScene.DetectPoint(ThisScene.Start);
            Point  ingoal= ThisScene.DetectPoint(ThisScene.Goal);

            Goal = new VertexData_State_Astar( ingoal.X, ingoal.Y, null);
            Start = new VertexData_State_Astar(instart.X, instart.Y, Goal);
            Graph.InitialVertex = new Vertex<VertexData_State_Astar, EdgeData_Heuristic>(Start);
        }

            void FindPath()
        {
            List<Vertex<VertexData_State_Astar, EdgeData_Heuristic>> OpenList = new List<Vertex<VertexData_State_Astar, EdgeData_Heuristic>>();
            List<Vertex<VertexData_State_Astar, EdgeData_Heuristic>> ClosedList = new List<Vertex<VertexData_State_Astar, EdgeData_Heuristic>>();
            AllNodesList = new List<Vertex<VertexData_State_Astar, EdgeData_Heuristic>>();

            bool found = false;
            OpenList.Add(Graph.InitialVertex);
            Vertex<VertexData_State_Astar, EdgeData_Heuristic> Unit = null;
            do
            {
                        double MinimalCost = Double.MaxValue;
                        int i = 0;
                        int idx = -1;
                        foreach (Vertex<VertexData_State_Astar, EdgeData_Heuristic> Item in OpenList)
                        {
                            if (Item.VertexData.Cost < MinimalCost)
                            {
                                MinimalCost = Item.VertexData.Cost;
                                idx = i;
                            }
                            i += 1;
                        }
                        Unit = OpenList[idx];
                        OpenList.RemoveAt(idx);
                       
                Vertex<VertexData_State_Astar, EdgeData_Heuristic> NextNode = Unit;
                ClosedList.Add(NextNode);
                
                List<VertexData_State_Astar> ListNewDataVertexex = NextNode.VertexData.Expand();
                foreach (VertexData_State_Astar v in ListNewDataVertexex)
                {
                    
                    if ((v.X < (ThisScene.MaxMapSize.X)) & (v.Y < (ThisScene.MaxMapSize.Y)) & (v.X > 0) & (v.Y > 0))
                    {
                        if (ThisScene.IsWalkable(NextNode.VertexData.X, NextNode.VertexData.Y, v.X, v.Y))
                        {
                            Vertex<VertexData_State_Astar, EdgeData_Heuristic> SolutionUnit = new Vertex<VertexData_State_Astar, EdgeData_Heuristic>(v);
                            if (v.Equals(Goal))
                            {
                                Solution = new List<Vertex<VertexData_State_Astar, EdgeData_Heuristic>>();
                                SolutionUnit.ConnectVertex(NextNode, null, false);
                                Solution.Add(SolutionUnit);
                                found = true;
                                break;

                            }
                            if (!IsInList(SolutionUnit, AllNodesList))   // neni li v open ani closed pak jej pridam do open
                            {
                                SolutionUnit.ConnectVertex(NextNode, null, false);
                                AllNodesList.Add(SolutionUnit);
                                OpenList.Add(SolutionUnit);
                            }
                        }
                    }
                }
             
            }
            
            while ((found == false) && (OpenList.Count > 0));

            OpenList.Clear();
            ClosedList.Clear();
            if (found == true)
            {
                ReconstructPath();   
            }
        }
       

        void ReconstructPath()
        { 
        if (Solution == null) { return; }
        Vertex<VertexData_State_Astar, EdgeData_Heuristic> Path = Solution[0].GetInputEdges[0].GetOutVertex;
            while (Path.GetInputEdges.Count  != 0)
            {
                Solution.Insert(0, Path);
                Path = Path.GetInputEdges[0].GetOutVertex;
            }
        }

        public List<Point> ReturnPoints()
        {
            List<Point> ListOfPoints = new List<Point>();
            FindPath();

            foreach (Vertex<VertexData_State_Astar, EdgeData_Heuristic> v in Solution)
            {
                ListOfPoints.Add(ThisScene.PointsOfHexagonalLattice[Convert.ToInt32(v.VertexData.X),Convert.ToInt32( v.VertexData.Y)]);
            }
            return ListOfPoints;
        }
        public List<Point> ReturnSearchedNodes()
        {
            List<Point> ListOfPoints = new List<Point>();
        

            foreach (Vertex<VertexData_State_Astar, EdgeData_Heuristic> v in AllNodesList)
            {
                ListOfPoints.Add(ThisScene.PointsOfHexagonalLattice[Convert.ToInt32(v.VertexData.X), Convert.ToInt32(v.VertexData.Y)]);
            }
            return ListOfPoints;
        }
    }
}

