﻿using System;
using System.Collections.Generic;
using System.Linq;
using SimulacniProstredi.Grafy;
using System.Drawing;
using System.Drawing.Drawing2D;
using SimulacniProstredi.Sceny;
using System.Text;

namespace SimulacniProstredi.Algoritmy
{
    class LPAStar
    {
       
        List<Vertex<VertexData_State_LPA, EdgeData_Heuristic>> CurrentNodesList=new List<Vertex<VertexData_State_LPA, EdgeData_Heuristic>>();
        List<Vertex<VertexData_State_LPA, EdgeData_Heuristic>> AllNodesList =new List<Vertex<VertexData_State_LPA, EdgeData_Heuristic>>();
        List<Vertex<VertexData_State_LPA, EdgeData_Heuristic>> ExpandedNodesList = new List<Vertex<VertexData_State_LPA, EdgeData_Heuristic>>();
        PriorityQueue_LPA Queue;
        StateSpace<VertexData_State_LPA, EdgeData_Heuristic> Graph = new StateSpace<VertexData_State_LPA, EdgeData_Heuristic>();
        VertexData_State_LPA Start;
        Scene ThisScene;
        List<Vertex<VertexData_State_LPA, EdgeData_Heuristic>> Solution = new List<Vertex<VertexData_State_LPA, EdgeData_Heuristic>>();
        public int NumberOfMathOperations = 0;
        Boolean FirstRun = true;
        Vertex<VertexData_State_LPA, EdgeData_Heuristic> GoalVertex;

        

        private bool IsInList(Vertex<VertexData_State_LPA, EdgeData_Heuristic> Vertex, List<Vertex<VertexData_State_LPA, EdgeData_Heuristic>> Vertexes)
        {
            foreach (Vertex<VertexData_State_LPA, EdgeData_Heuristic> v in Vertexes)
            {
                if (Vertex.VertexData.Equals(v.VertexData))
                { return true; }
            }
            return false;
        }

        private bool IsInEdgeList(Edge<VertexData_State_LPA, EdgeData_Heuristic> edge, List<Edge<VertexData_State_LPA, EdgeData_Heuristic>> Edges)
        {
            foreach (Edge<VertexData_State_LPA, EdgeData_Heuristic> e in Edges)
            {
                if (edge.Equals(e))
                { return true; }
            }
            return false;
        }

        public LPAStar(Scene InScene) 
        {
            ThisScene = InScene;
            Point instart= ThisScene.DetectPoint(ThisScene.Start);
            Point  ingoal= ThisScene.DetectPoint(ThisScene.Goal);

            GoalVertex = new Vertex<VertexData_State_LPA, EdgeData_Heuristic>(new VertexData_State_LPA(ingoal.X, ingoal.Y));
            Start = new VertexData_State_LPA(instart.X, instart.Y, GoalVertex.VertexData);
            Graph.InitialVertex = new Vertex<VertexData_State_LPA, EdgeData_Heuristic>(Start);
        }

        void Initialize()
        {
            Queue = new PriorityQueue_LPA();
            Graph.InitialVertex.VertexData.InitializeVertexData();
            Graph.InitialVertex.VertexData.Setrhs = 0;
            Queue.Insert(Graph.InitialVertex,QueueUnit_LPA.ToPriority(Graph.InitialVertex.VertexData.Geth,0));
            GoalVertex.VertexData.InitializeVertexData();
        }

        void UpdateVertex(Vertex<VertexData_State_LPA, EdgeData_Heuristic> inVertex)
        { 
          if (!(inVertex.VertexData.Equals(Start)) )
          {
              inVertex.VertexData.Setrhs=inVertex.GetInputEdges[0].GetOutVertex.VertexData.Getg + inVertex.GetInputEdges[0].EdgeData.GetCost; //Pokud nepojede smazat, pokud pojede zlidstit
              
              foreach (Edge<VertexData_State_LPA,EdgeData_Heuristic> e in inVertex.GetInputEdges)
            {
                 int rhs = e.GetOutVertex.VertexData.Getg + e.EdgeData.GetCost; //zjistit proc se mně někdy otáčí hodnota rhs a g

                 if (rhs < inVertex.VertexData.Getrhs) inVertex.VertexData.Setrhs  = rhs;
            }
           
           }
          if ((Queue.IsInQueue(inVertex))&(inVertex.VertexData.Getg != inVertex.VertexData.Getrhs)) Queue.Update(inVertex,inVertex.VertexData.CalculateKey);
          else if((Queue.IsInQueue(inVertex)) & (inVertex.VertexData.Getg == inVertex.VertexData.Getrhs)) Queue.Remove(inVertex);
          else if (!(Queue.IsInQueue(inVertex)) & (inVertex.VertexData.Getg != inVertex.VertexData.Getrhs) & !IsInList(inVertex, ExpandedNodesList))
          { ExpandedNodesList.Add(inVertex); Queue.Insert(inVertex, inVertex.VertexData.CalculateKey);
          CurrentNodesList.Add(inVertex);
          }
    
        }

        void ComputeShrotestPath()
        {
            int ComputeRuns = 0;
              while(!GoalVertex.VertexData.HaveSmallerPriorityThanOther(Queue.TopKey)| (GoalVertex.VertexData.Getrhs!=GoalVertex.VertexData.Getg ))
            {
                ComputeRuns = ComputeRuns + 1;
                Vertex<VertexData_State_LPA, EdgeData_Heuristic> Vertex = Queue.Pop();
                ManageSuccesors(Vertex);
                
                  if (Vertex.VertexData.Getg>Vertex.VertexData.Getrhs)
                  {
                      Vertex.VertexData.Getg = Vertex.VertexData.Getrhs;
                      foreach (Edge<VertexData_State_LPA, EdgeData_Heuristic> e in Vertex.GetOutputEdges)
                      {
                          UpdateVertex(e.GetIntoVertex);
                      }
                  }
                  else
                  {
                      Vertex.VertexData.Getg = int.MaxValue / 3;
                      foreach (Edge<VertexData_State_LPA, EdgeData_Heuristic> e in Vertex.GetOutputEdges)
                      {
                          UpdateVertex(e.GetIntoVertex);
                      }
                      UpdateVertex(Vertex);
                  }
            }
              Solution = new List<Vertex<VertexData_State_LPA, EdgeData_Heuristic>>();
              Solution.Add(GoalVertex);
              ReconstructPath();
        }
        void ReconstructPath()
        {//jit po nejniším rsh
            if (Solution == null) { return; }

            Vertex<VertexData_State_LPA, EdgeData_Heuristic> Path = Solution[0];

            while (Path.GetInputEdges.Count != 0)
            {
                int minRhs = Path.GetInputEdges[0].GetOutVertex.VertexData.Getrhs;  
            int minEdgeNumber=0;;
            for (int i = 1; i < Path.GetInputEdges.Count(); i++)
            {
               
                if (Path.GetInputEdges[i].GetOutVertex.VertexData.Getrhs < minRhs) { minRhs = Path.GetInputEdges[i].GetOutVertex.VertexData.Getrhs; minEdgeNumber = i; }
                
            }
                Solution.Insert(0, Path);
                Path = Path.GetInputEdges[minEdgeNumber].GetOutVertex;
            }
        }
        void ManageSuccesors(Vertex<VertexData_State_LPA, EdgeData_Heuristic> Vertex)
        {
            if (Vertex.GetOutputEdges.Count() == 0)
            {
                List<VertexData_State_LPA> ListNewDataVertexex = Vertex.VertexData.Expand();

                foreach (VertexData_State_LPA v in ListNewDataVertexex)
                {

                    if ((v.X < (ThisScene.MaxMapSize.X)) & (v.Y < (ThisScene.MaxMapSize.Y)) & (v.X > 0) & (v.Y > 0))
                    {
                        Vertex<VertexData_State_LPA, EdgeData_Heuristic> Unit = new Vertex<VertexData_State_LPA, EdgeData_Heuristic>(v);
                        Vertex<VertexData_State_LPA, EdgeData_Heuristic> UnitAlredyInList = AllNodesList.Find(delegate(Vertex<VertexData_State_LPA, EdgeData_Heuristic> AlreadyExpanded) { return AlreadyExpanded.VertexData.Equals(v); });

                        if (ThisScene.IsWalkable(Vertex.VertexData.X, Vertex.VertexData.Y, v.X, v.Y))
                        {

                            if (UnitAlredyInList != null)
                            {
                                Boolean FoundIt = false;
                                foreach (Edge<VertexData_State_LPA, EdgeData_Heuristic> e in UnitAlredyInList.GetInputEdges)
                                {
                                    if (e.GetOutVertex.VertexData.Equals(Vertex.VertexData)) FoundIt = true;
                                }
                                if (!FoundIt) { UnitAlredyInList.ConnectVertex(Vertex, new EdgeData_Heuristic(1), false); }
                            }
                            else
                            {
                                Unit.ConnectVertex(Vertex, new EdgeData_Heuristic(1), false);
                                Unit.VertexData.InitializeVertexData();
                                if (Unit.VertexData.Equals(GoalVertex.VertexData))
                                { GoalVertex = Unit; }

                                AllNodesList.Add(Unit);
                            }

                        }
                        else
                        {
                            if (UnitAlredyInList != null)
                            {
                                Boolean FoundIt = false;
                                foreach (Edge<VertexData_State_LPA, EdgeData_Heuristic> e in UnitAlredyInList.GetInputEdges)
                                {
                                    if (e.GetOutVertex.VertexData.Equals(Vertex.VertexData)) FoundIt = true;
                                }
                                if (!FoundIt) { UnitAlredyInList.ConnectVertex(Vertex, new EdgeData_Heuristic(int.MaxValue / 2), false);


                                }
                            }
                            else
                            {
                                Unit.ConnectVertex(Vertex, new EdgeData_Heuristic(int.MaxValue / 2), false);
                                AllNodesList.Add(Unit);
                                Unit.VertexData.InitializeVertexData();
                            }
                        }
                    }
                }
            }
        }

    public void UseLPA_FirstTime()
        {
            Initialize();

            ComputeShrotestPath();

        }
       public void UseLPA()
        {
            FirstRun = false;

         GoalVertex.VertexData .InitializeVertexDataForSecondSearch();//mozna zakomentovat, potom co pretedam create vertexes na managesuccesors
         List<Edge<VertexData_State_LPA, EdgeData_Heuristic>> ListOfEdges, ListOfRemovedEdges = new List<Edge<VertexData_State_LPA, EdgeData_Heuristic>>();
            ListOfEdges = CheckEdge();
            ListOfRemovedEdges = CheckEdgeForRemovedEdges();
           
           ListOfEdges.ForEach(delegate(Edge<VertexData_State_LPA, EdgeData_Heuristic> e)
            {
               e.EdgeData.SetCost = (int.MaxValue / 2);
            });
           ListOfRemovedEdges.ForEach(delegate(Edge<VertexData_State_LPA, EdgeData_Heuristic> e)
           {
               e.EdgeData.SetCost = 1;
           });
           ListOfEdges.AddRange(ListOfRemovedEdges);
           ListOfEdges.ForEach(delegate(Edge<VertexData_State_LPA, EdgeData_Heuristic> e)
           {

                   UpdateVertex(e.GetIntoVertex);

           });
           ExpandedNodesList=new List<Vertex<VertexData_State_LPA, EdgeData_Heuristic>>();//huh?

           
           AllNodesList.ForEach(delegate(Vertex<VertexData_State_LPA, EdgeData_Heuristic> v)
           {
               if (v.VertexData.Getg != v.VertexData.Getrhs)
               {
                   if (!Queue.IsInQueue(v)) 
                       Queue.Insert(v, v.VertexData.CalculateKey);
               }
           });
          
           CurrentNodesList = new List<Vertex<VertexData_State_LPA, EdgeData_Heuristic>>();
           ComputeShrotestPath();
        }

       List<Edge<VertexData_State_LPA, EdgeData_Heuristic>> CheckEdgeForRemovedEdges()
       {

           List<Edge<VertexData_State_LPA, EdgeData_Heuristic>> ListOfEdges = new List<Edge<VertexData_State_LPA, EdgeData_Heuristic>>();

           AllNodesList.ForEach(delegate(Vertex<VertexData_State_LPA, EdgeData_Heuristic> v)
           {
               foreach (Edge<VertexData_State_LPA, EdgeData_Heuristic> e in v.GetOutputEdges)
               {

                   if (e.EdgeData.GetCost == (int.MaxValue / 2))
                   {
                       if (ThisScene.IsWalkable(e.GetOutVertex.VertexData.X, e.GetOutVertex.VertexData.Y, e.GetIntoVertex.VertexData.X, e.GetIntoVertex.VertexData.Y))
                       {

                           if (!IsInEdgeList(e, ListOfEdges)) ListOfEdges.Add(e);
                       }
                   }
               }
           });
           return ListOfEdges;
       } 


       List<Edge<VertexData_State_LPA, EdgeData_Heuristic>> CheckEdge()
       {

           List<Edge<VertexData_State_LPA, EdgeData_Heuristic>> ListOfEdges = new List<Edge<VertexData_State_LPA, EdgeData_Heuristic>>();

           AllNodesList.ForEach(delegate(Vertex<VertexData_State_LPA, EdgeData_Heuristic> v)
           {
              // foreach (Edge<VertexData_State_LPA, EdgeData_Heuristic> e in v.GetInputEdges)
                   foreach (Edge<VertexData_State_LPA, EdgeData_Heuristic> e in v.GetOutputEdges)
               {

                   if (!ThisScene.IsWalkable(e.GetOutVertex.VertexData.X, e.GetOutVertex.VertexData.Y, e.GetIntoVertex.VertexData.X, e.GetIntoVertex.VertexData.Y))
                   {
                       if (e.EdgeData.GetCost < (int.MaxValue / 2))
                       {
                           if (!IsInEdgeList(e, ListOfEdges)) ListOfEdges.Add(e);
                       }
                   }

               }


           });
           return ListOfEdges;

       }

        public List<Point> ReturnPoints()
        {
            List<Point> ListOfPoints = new List<Point>();
              foreach (Vertex<VertexData_State_LPA, EdgeData_Heuristic> v in Solution)
            {
                ListOfPoints.Add(ThisScene.PointsOfHexagonalLattice[Convert.ToInt32(v.VertexData.X), Convert.ToInt32(v.VertexData.Y)]);
            }
            return ListOfPoints;
        }
        public List<Point> ReturnSearchedNodes()
        {
            List<Point> ListOfPoints = new List<Point>();
            foreach (Vertex<VertexData_State_LPA, EdgeData_Heuristic> v in AllNodesList)
            {
                ListOfPoints.Add(ThisScene.PointsOfHexagonalLattice[Convert.ToInt32(v.VertexData.X), Convert.ToInt32(v.VertexData.Y)]);
            }
            return ListOfPoints;
        }
        public List<Point> ReturnNewlySearchedNodes()
        {
            List<Point> ListOfPoints = new List<Point>();
            foreach (Vertex<VertexData_State_LPA, EdgeData_Heuristic> v in CurrentNodesList)
            {
                ListOfPoints.Add(ThisScene.PointsOfHexagonalLattice[Convert.ToInt32(v.VertexData.X), Convert.ToInt32(v.VertexData.Y)]);
            }
            return ListOfPoints;
        }
    }
}
