﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace SimulacniProstredi
{
   public class Vertex<inDataV, inDataE> : IEquatable<Vertex<inDataV, inDataE>>, IComparable<Vertex<inDataV, inDataE>>
              where inDataV : IEquatable<inDataV>, IComparable<inDataV>
              where inDataE : IEquatable<inDataE>, IComparable<inDataE>
    {
    int id;
    public int ID
    {
        get { return id; }
    }
    public bool searched = false;
    public inDataV VertexData;
    private static List<Vertex<inDataV, inDataE>> SearchedVertexes = new List<Vertex<inDataV, inDataE>>();
    private List<Edge<inDataV, inDataE>> InputEdges = new List<Edge<inDataV, inDataE>>();
    public  List<Edge<inDataV, inDataE>> GetInputEdges
    { get { return InputEdges; } }
    private List<Edge<inDataV, inDataE>> OutputEdges = new List<Edge<inDataV, inDataE>>();
    public List<Edge<inDataV, inDataE>> GetOutputEdges
    { get { return OutputEdges; } }
    static int MaxVertexId = 0;
    int GetVertexId()
    {
        MaxVertexId = MaxVertexId + 1;
        return MaxVertexId;
    }
  public  Vertex(inDataV inDataV)
    {
        this.id = GetVertexId();
        this.VertexData = inDataV;
    }

   public void AddVertex(inDataV inDataV, inDataE inDataE)
    {
        Vertex<inDataV, inDataE> NewVertex = new Vertex<inDataV, inDataE>(inDataV);
        new Edge<inDataV, inDataE>(inDataE,this ,NewVertex);
    }

   public void AddPrevVertex(inDataV inDataV, inDataE inDataE)
   {
       Vertex<inDataV, inDataE> NewVertex = new Vertex<inDataV, inDataE>(inDataV);
       new Edge<inDataV, inDataE>(inDataE, NewVertex, this);

   }

   public void ConnectVertex(Vertex<inDataV, inDataE> VertexToConnect, inDataE inDataE, Boolean IsItNextVertex)
   { 
    if (IsItNextVertex == false) 
    {
        new Edge<inDataV, inDataE>(inDataE, VertexToConnect, this);
    }
    else 
    {
        new Edge<inDataV, inDataE>(inDataE, this, VertexToConnect);
    }
    }

   public void RemoveVertex()
   {
       if (this.GetOutputEdges.Count() != 0)
       {
           foreach (Edge<inDataV, inDataE> Edges in this.GetOutputEdges)
           {
               Edges.GetOutVertex.OutputEdges.Remove(Edges);
               if (Edges.GetIntoVertex.InputEdges.Count == 1)
               {
                   Edges.GetIntoVertex.RemoveVertex();
               }
               if (Edges.GetIntoVertex != null)
                   Edges.GetIntoVertex.InputEdges.Remove(Edges);
               Edges.RemoveEdge();
               if (this.GetOutputEdges != null) break;
           }

       }
       else
       {
           int NumberOfRemovals = (this.GetInputEdges.Count());
           for (int i = 0; i < NumberOfRemovals; i++)
           {
               this.GetInputEdges[0].GetOutVertex.OutputEdges.Remove(this.GetInputEdges[0]);
               this.GetInputEdges.Remove(this.GetInputEdges[0]);
           }
           
       }
           if ((this.GetInputEdges.Count == 1) && (this.GetOutputEdges.Count == 0))
           {
               this.GetInputEdges[0].GetOutVertex.GetOutputEdges.Remove(this.GetInputEdges[0]);
               this.InputEdges[0].RemoveEdge();
           }
       
   }


   public void AddNextEdge(inDataE inDataE, Vertex<inDataV, inDataE> TargetVertex)
   {
       new Edge<inDataV, inDataE>(inDataE, this, TargetVertex);
   }



   public bool FindIt(int inDepth, Vertex<inDataV, inDataE> WantedVertex, List<Edge<inDataV, inDataE>> Path)
   {
       if (this.searched)
       {
           return false;

       }
       searched = true;
       SearchedVertexes.Add(this);

       if (this.Equals(WantedVertex))
       {
           Path = new List<Edge<inDataV, inDataE>>();
           return true;
       }

       foreach (Edge<inDataV, inDataE> Edges in this.OutputEdges)
       {

           if (Edges.GetIntoVertex.FindIt(inDepth - 1, WantedVertex, Path))
           {
               Path.Insert(0, Edges);
               return true;
           }
       }
       return false;
   }
   public bool FindIt(Vertex<inDataV, inDataE> WantedVertex, List<Edge<inDataV, inDataE>> Path)
   {
       if (this.searched)
       {
           return false;

       }
       searched = true;
       SearchedVertexes.Add(this);

       if (this.Equals(WantedVertex))
       {
           Path = new List<Edge<inDataV, inDataE>>();
           return true;
       }

       foreach (Edge<inDataV, inDataE> hrany in this.OutputEdges)
       {

           if (hrany.GetIntoVertex.FindIt(WantedVertex, Path))
           {
               Path.Insert(0, hrany);
               return true;
           }
       }
       return false;
   }
   public bool FindItBackwards(Vertex<inDataV, inDataE> WantedVertex, List<Edge<inDataV, inDataE>> Path)
   {
       if (this.searched)
       {
           return false;

       }
       searched = true;
       SearchedVertexes.Add(this);

       if (this.Equals(WantedVertex))
       {
           Path = new List<Edge<inDataV, inDataE>>();
           return true;
       }

       foreach (Edge<inDataV, inDataE> Edges in this.InputEdges)
       {

           if (Edges.GetOutVertex.FindItBackwards(WantedVertex, Path))
           {
               Path.Insert(0, Edges);
               return true;
           }
       }
       return false;
   }

   public bool FindItdata(inDataV WantedData,List<Edge<inDataV, inDataE>> Patch)
   {
       if (this.searched)
       {
           return false;

       }
       searched = true;
       SearchedVertexes.Add(this);

       if (this.VertexData.Equals(WantedData))
       {
           Patch = new List<Edge<inDataV, inDataE>>();
           return true;
       }
    
       foreach (Edge<inDataV, inDataE> Edges in this.OutputEdges)
       {

           if (Edges.GetIntoVertex.FindItdata( WantedData, Patch))
           {
               Patch.Insert(0, Edges);
               return true;
           }
       }

       return false;
   }
   public static void ClearSearchedVertexes()
   {
       foreach (Vertex<inDataV, inDataE> Vertexes in SearchedVertexes)
       {
           Vertexes.searched = false;
       }
   }

   public bool FindItId(int WantedId, List<Edge<inDataV, inDataE>> Patch)
   {
       if (this.searched)
       {
           return false;

       }
       searched = true;
       SearchedVertexes.Add(this);

       if (this.ID.Equals(WantedId))
       {
           Patch = new List<Edge<inDataV, inDataE>>();
           return true;
       }

       foreach (Edge<inDataV, inDataE> Edges in this.OutputEdges)
       {

           if (Edges.GetIntoVertex.FindItId(WantedId, Patch))
           {
               Patch.Insert(0, Edges);
               return true;
           }
       }

       return false;
   }

   #region IComparable<Uzel<vdataU,vdataH>> Members

   public int CompareTo(Vertex<inDataV, inDataE> other)
   {
       return (id - other.ID);
   }

   #endregion

   #region IEquatable<Uzel<vdataU,vdataH>> Members

   public bool Equals(Vertex<inDataV, inDataE> other)
   {
       if (this.id == other.id) { return true; }
       else { return false; };
   }

   #endregion
   public override string ToString()
   {

       string description;
       description = String.Format("Id: {0}, {1}", id, VertexData);
       return description;
      
    }
   public Edge<inDataV, inDataE> FindEdge_VertexId(int vid)
   {
       foreach (Edge<inDataV, inDataE> Edges in this.GetOutputEdges)
       {
           if (Edges.GetOutVertex.ID.Equals(vid) == true)
           {
               return Edges;
           }
           else { return FindEdge_VertexId(vid); }
       }
       return null;
   }

   public Edge<inDataV, inDataE> FindEdge_EdgeId(int vid)
   {
       foreach (Edge<inDataV, inDataE> Edges in this.GetOutputEdges)
       {
           if (Edges.ID.Equals(vid) == true)
           {
               return Edges;
           }
           else { return FindEdge_EdgeId(vid); }
       }
       return null;
   }

   public bool FindEdge_EdgeData(inDataE vdata, int depth, Edge<inDataV, inDataE> Edge)
   {
           foreach (Edge<inDataV, inDataE> Edges in this.GetOutputEdges)
           {
               if (Edges.EdgeData.Equals(vdata) == true)
               {
                   Edge = Edges;
                   return true;
               }
               else if (depth >0) { return FindEdge_EdgeData(vdata, depth - 1, Edge); }
           }
       
       return false;
   }
   public bool FindEdge_EdgeData_backwards(inDataE vdata, int depth, Edge<inDataV, inDataE> Edge)
   {
       foreach (Edge<inDataV, inDataE> Edges in this.GetInputEdges)
       {
           if (Edges.EdgeData.Equals(vdata) == true)
           {
               Edge = Edges;
               return true;
           }
           else if (depth > 0) { return FindEdge_EdgeData_backwards(vdata, depth - 1, Edge); }
       }

       return false;
   }
}
}
